#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <string.h>

#include "lcd.h"

#define SWAP(x,y) {x ^= y; y ^= x; x ^= y; }
#define abs(x, type) \
(__extension__({                \
	type x_ = (x); \
	x_ > 0 ? x : -x; \
}))

#define LCD_SELECT() {LCD_PORT &= ~_BV(LCD_CS);}
#define LCD_DESELECT() {LCD_PORT |= _BV(LCD_CS);}

#define SPI_SEND(b) {\
    /*SPCR =  (1<<SPE) | (1<<MSTR);*/ \
    SPDR = b;\
    while (!(SPSR & _BV(SPIF)));  \
}

#define RGB(r,g,b) ( ((r & 0x07) << 5) | ((g & 0x07) << 2) | ((b & 0x03)))

/*
 * Нашол! В инициализации код $7f3f заменяем на $7f1f ,получаем 256 цветов. Передача команд таже , дата не 16бит один пиксель , а 16бит два пикселя.
 */

uint8_t PROGMEM INIT1[] = {
	0xFD, 0xFD, 0xFD, 0xFD, 0xEF, 0x00, 0xEE, 0x04, 0x1B, 0x04, 0xFE, 0xFE,
	//0xFE, 0xFE, 0xEF, 0x90, 0x4A, 0x04, 0x7F, 0x3F, 0xEE, 0x04, 0x43, 0x06
	0xFE, 0xFE, 0xEF, 0x90, 0x4A, 0x04, 0x7F, 0x1F, 0xEE, 0x04, 0x43, 0x06
};

uint8_t PROGMEM INIT2[] = {
	0xEF, 0x90, 0x09, 0x83, 0x08, 0x00, 0x0B, 0xAF, 0x0A, 0x00,
	0x05, 0x00, 0x06, 0x00, 0x07, 0x00, 0xEF, 0x00, 0xEE, 0x0C,
	0xEF, 0x90, 0x00, 0x80, 0xEF, 0xB0, 0x49, 0x02, 0xEF, 0x00,
	0x7F, 0x01, 0xE1, 0x81, 0xE2, 0x02, 0xE2, 0x76, 0xE1, 0x83
};

uint8_t PROGMEM INIT3[] = { 0x80, 0x01 };

static void lcd_wrcmd(uint8_t cmd);
//static void lcd_wrdata(uint8_t data);

static uint8_t clip_x0 = 0;
static uint8_t clip_y0 = 0;
static uint8_t clip_x1 = LCD_WIDTH - 1;
static uint8_t clip_y1 = LCD_HEIGHT - 1;

void lcd_clip(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
	clip_x0 = x0;
	clip_y0 = y0;
	clip_x1 = x1;
	clip_y1 = y1;
}


static void port_init_io(void)
{

	LCD_DDR &= ~_BV(LCD_MISO);
	LCD_DDR |= _BV(LCD_CS) | _BV(LCD_SCK) | _BV(LCD_MOSI) | _BV(SPI_SS);
	LCD_DDR2 |= _BV(LCD_RS) | _BV(LCD_RESET);

	LCD_PORT &= ~(_BV(LCD_SCK) | _BV(SPI_SS));
	LCD_PORT2 &= ~(_BV(LCD_RESET));
	LCD_PORT |= _BV(LCD_CS) | _BV(LCD_SCK) | _BV(LCD_MISO);
	LCD_PORT2 |= _BV(LCD_RS);


	SPCR = _BV(MSTR) | _BV(SPE);
	SPSR = 0x01;

	_delay_ms(1);
	LCD_PORT2 |= _BV(LCD_RESET);	/* release reset */
	(void) SPSR;
}


void wrtab(uint8_t * table, uint8_t len)
{
	while (len) {
		lcd_wrcmd(pgm_read_byte(table));
		table++;
		len--;
	}
}


void lcd_init(void)
{
	LCD_DESELECT();
	port_init_io();

	wrtab(INIT1, 24);
	_delay_ms(7);

	wrtab(INIT2, 40);
	_delay_ms(10);
	_delay_ms(10);
	_delay_ms(10);
	_delay_ms(10);
	_delay_ms(10);

	wrtab(INIT3, 2);

	lcd_wrcmd(0xEF);
	lcd_wrcmd(0x90);
	lcd_wrcmd(0x00);
	lcd_wrcmd(0x00);

	lcd_clrscr();
}


void lcd_clrscr(void)
{
	/* send window write start sequence */
	lcd_wrcmd(0xEF);
	lcd_wrcmd(0x90);
	lcd_wrcmd(0x05);
	lcd_wrcmd(0x00);
	lcd_wrcmd(0x06);
	lcd_wrcmd(0x00);
	lcd_wrcmd(0x07);
	lcd_wrcmd(0x00);


	/* now 132*176*2 bytes are to be transmitted */
	lcd_wrdat16xN(0x00, 0xB580U >> 1);
}

static void lcd_wrcmd(uint8_t cmd)
{
	LCD_SELECT();
	LCD_PORT2 |= _BV(LCD_RS);	/* set RS line */
	SPI_SEND(cmd);
	LCD_DESELECT();
}


void lcd_wrdata(uint8_t data)
{
  	LCD_SELECT();
    LCD_PORT2 &= ~_BV(LCD_RS);
    SPI_SEND(data);
	LCD_DESELECT();
}

void lcd_wrcmd16(uint16_t dat)
{
	LCD_SELECT();
	LCD_PORT2 |= _BV(LCD_RS);	/* set RS line */
	SPI_SEND((dat >> 8));
	SPI_SEND(dat);
	LCD_DESELECT();
}

void lcd_wrdat16(uint16_t dat)
{
	LCD_SELECT();
	LCD_PORT2 &= ~_BV(LCD_RS);
	SPI_SEND((dat >> 8));
	SPI_SEND(dat);
	LCD_DESELECT();
}

void lcd_wrdat16xN(uint16_t dat, uint16_t n)
{
	LCD_SELECT();
	LCD_PORT2 &= ~_BV(LCD_RS);
	while (n) {
		SPI_SEND((dat >> 8));
		SPI_SEND(dat);
		n--;
	}
	LCD_DESELECT();
}

void lcd_wrdat8xN(uint8_t dat, uint16_t n)
{
	LCD_SELECT();
	LCD_PORT2 &= ~_BV(LCD_RS);
	while (n) {
		SPI_SEND(dat);
		n--;
	}
	LCD_DESELECT();
}


static uint8_t window_x, window_y, window_w, window_h;

void lcd_window(uint8_t x, uint8_t y, uint8_t w, uint8_t h)
{
	LCD_SELECT();
	LCD_PORT2 |= _BV(LCD_RS);
	SPI_SEND(0x08);
	SPI_SEND(x);
	SPI_SEND(0x09);
	SPI_SEND(x + w - 1);
	SPI_SEND(0x0A);
	SPI_SEND(y);
	SPI_SEND(0x0B);
	SPI_SEND(y + h - 1);
	SPI_SEND(0x06);
	SPI_SEND(x);
	SPI_SEND(0x07);
	SPI_SEND(y);
	LCD_DESELECT();
	window_x = x;
	window_y = y;
	window_w = w;
	window_h = h;
}



void lcd_fill_window(uint8_t color)
{
	uint16_t i;
	LCD_SELECT();
	LCD_PORT2 |= _BV(LCD_RS);
	SPI_SEND(0x06);
	SPI_SEND(window_x);
	SPI_SEND(0x07);
	SPI_SEND(window_y);
	LCD_PORT2 &= ~_BV(LCD_RS);
	for (i = 0; i < window_w * window_h; i++) {
		SPI_SEND(color);
	}
	LCD_DESELECT();
}

void lcd_pos(uint8_t x, uint8_t y)
{
	LCD_SELECT();
	LCD_PORT2 |= _BV(LCD_RS);
	SPI_SEND(0x06);
	SPI_SEND(x);
	SPI_SEND(0x07);
	SPI_SEND(y);
	LCD_DESELECT();
}

uint8_t fgColor = 0xFF;
uint8_t bgColor = 0x80;

void lcd_char(char c, uint8_t x, uint8_t y)
{
	uint8_t i;
	lcd_window(x, y, FONT_W, FONT_H);
	for (i = 0; i < 12; i++) {
		lcd_pos(x, y);
		LCD_SELECT();
		LCD_PORT2 &= ~_BV(LCD_RS);
		uint8_t j;
		uint8_t bits = pgm_read_byte(&(ascii_tab[(uint8_t) c][i]));
		for (j = 0; j < 8; j++) {
			SPI_SEND((bits & 0x80) ? fgColor : bgColor);
			bits <<= 1;
		}
		LCD_DESELECT();
		y++;
	}
}

void lcd_string(char *str, uint8_t x, uint8_t y, uint8_t w, uint8_t h, uint8_t rom)
{
	uint8_t i;
	lcd_window(x, y, w, h);

	char *s = str;
	char *next_line = str;

	char c = rom ? pgm_read_byte(s) : *s;

	while (c) {
		for (i = 0; i < 12; i++) {
			s = next_line;
			lcd_pos(x, y);
			LCD_SELECT();
			LCD_PORT2 &= ~_BV(LCD_RS);
			y++;
			c = rom ? pgm_read_byte(s) : *s;
			while (c) {
				uint8_t j;
				uint8_t bits = pgm_read_byte(&(ascii_tab[(uint8_t) c][i]));
				for (j = 0; j < 8; j++) {
					SPI_SEND((bits & 0x80) ? fgColor : bgColor);
					bits <<= 1;
				}
				s++;
				c = rom ? pgm_read_byte(s) : *s;
				if (c  == '\n') {
					if (i == 11) {
						next_line = s + 1;
					}
					break;
				}
			}
			LCD_DESELECT();
		}
	}
}

void lcd_start_update(void)
{
	lcd_wrcmd16(0x0080);
}

void lcd_end_update(void)
{
	lcd_wrcmd16(0x0000);
}

void lcd_rect(uint8_t x, uint8_t y, uint8_t w, uint8_t h, uint8_t color)
{
	lcd_window(x, y, w, 1);
	lcd_fill_window(color);
	lcd_window(x + w - 1, y, 1, h);
	lcd_fill_window(color);
	lcd_window(x, y + h - 1, w, 1);
	lcd_fill_window(color);
	lcd_window(x, y, 1, h);
	lcd_fill_window(color);
}

void lcd_line(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1,
			  uint8_t color)
{
	if (x0 == x1) {
		if (y1 < y0) {
			SWAP(y0, y1);
		}
		lcd_window(x0, y0, 1, y1 - y0 + 1);
		lcd_fill_window(color);
		return;
	}
	if (y0 == y1) {
		if (x1 < x0) {
			SWAP(x0, x1);
		}
		lcd_window(x0, y0, x1 - x0 + 1, 1);
		lcd_fill_window(color);
		return;
	}
	int16_t Dx = x1 - x0;
	int16_t Dy = y1 - y0;
	uint8_t steep = (abs(Dy, int16_t) >= abs(Dx, int16_t));
	if (steep) {
		SWAP(x0, y0);
		SWAP(x1, y1);
		// recompute Dx, Dy after swap
		Dx = x1 - x0;
		Dy = y1 - y0;
	}
	lcd_window(0, 0, LCD_WIDTH - 1, LCD_HEIGHT - 1);
	int16_t xstep = 1;
	if (Dx < 0) {
		xstep = -1;
		Dx = -Dx;
	}
	int16_t ystep = 1;
	if (Dy < 0) {
		ystep = -1;
		Dy = -Dy;
	}
	int16_t TwoDy = 2 * Dy;
	int16_t TwoDyTwoDx = TwoDy - 2 * Dx;	// 2*Dy - 2*Dx
	int16_t E = TwoDy - Dx;		//2*Dy - Dx
	uint8_t y = y0;
	uint8_t xDraw, yDraw;
	uint8_t x;
	for (x = x0; x != x1; x += xstep) {
		if (steep) {
			xDraw = y;
			yDraw = x;
		} else {
			xDraw = x;
			yDraw = y;
		}
		// plot
		//plot(xDraw, yDraw);
		LCD_SELECT();
		LCD_PORT2 |= _BV(LCD_RS);
		SPI_SEND(0x06);
		SPI_SEND((uint8_t) xDraw);
		SPI_SEND(0x07);
		SPI_SEND((uint8_t) yDraw);
		LCD_DESELECT();
		LCD_SELECT();
		LCD_PORT2 &= ~_BV(LCD_RS);
		SPI_SEND(color);
		LCD_DESELECT();
		// next
		if (E > 0) {
			E += TwoDyTwoDx;	//E += 2*Dy - 2*Dx;
			y = y + ystep;
		} else {
			E += TwoDy;			//E += 2*Dy;
		}
	}
}


static inline void setPixel(uint8_t x, uint8_t y, uint8_t color)
{
	LCD_SELECT();
	LCD_PORT2 |= _BV(LCD_RS);
	SPI_SEND(0x06);
	SPI_SEND(x);
	SPI_SEND(0x07);
	SPI_SEND(y);
	LCD_DESELECT();
	LCD_SELECT();
	LCD_PORT2 &= ~_BV(LCD_RS);
	SPI_SEND(color);
	LCD_DESELECT();
}

void lcd_circle(uint8_t x0, uint8_t y0, uint8_t radius, uint8_t color)
{
	int16_t f = 1 - radius;
	int16_t ddF_x = 1;
	int16_t ddF_y = -2 * radius;
	uint8_t x = 0;
	uint8_t y = radius;

	setPixel(x0, y0 + radius, color);
	setPixel(x0, y0 - radius, color);
	setPixel(x0 + radius, y0, color);
	setPixel(x0 - radius, y0, color);

	while (x < y) {
		if (f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x;
		setPixel(x0 + x, y0 + y, color);
		setPixel(x0 - x, y0 + y, color);
		setPixel(x0 + x, y0 - y, color);
		setPixel(x0 - x, y0 - y, color);
		setPixel(x0 + y, y0 + x, color);
		setPixel(x0 - y, y0 + x, color);
		setPixel(x0 + y, y0 - x, color);
		setPixel(x0 - y, y0 - x, color);
	}
}



void lcd_bitmap_mono(uint8_t* bitmap, uint8_t x, uint8_t y, uint8_t bg, uint8_t fg)
{
	uint8_t byte = 0;
	uint8_t bit = 0;
	uint8_t w = pgm_read_byte(bitmap); bitmap++;
	uint8_t h = pgm_read_byte(bitmap); bitmap++;
	lcd_window(x, y, w, h);
	LCD_SELECT();
	LCD_PORT2 &= ~_BV(LCD_RS);
	for (; h; h--) {
		uint8_t w_;
		for (w_ = w; w_; w_--) {
			if (!bit) {
				byte = pgm_read_byte(bitmap); bitmap++;
			}
			if (byte & 0x80) {
				SPI_SEND(fg);
			} else {
				SPI_SEND(bg);
			}
			byte <<= 1;
			bit = (bit + 1) & 0x07;
		}
	}
	LCD_DESELECT();
}

